#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <recarray.h>
#include <tabutil.h>
#include <utility.h>
#include <lffiles.h>

#include "at1.h"
#include "atlib.h"

#define ALIAS_TCS	200
#define FINE_SOSPENSIONE "FS"

// nomi campi maschera
#include "at1300a.h"
          
// nomi dei campi
#include "soggetti.h"
#include "contsan.h"
#include "sezioni.h"

class TSbloccoSospesi : public TPrintapp
{
	TMask*						_msk;
	TRelation*   			_rel;
	TLocalisamfile*		_contsan;
	TRecord_array* 		_scontrolli;
	int 							_cur;	
	TDate							_data_sblocco; 			// data scadenza sospensione fino a _data_sblocco
	TDate							_data_stampa;
	TString16					_tiposo;						// tipo sospesione da sbloccare
	TString16					_codsez, _codsot;
	bool							_definitiva;
  TParagraph_string _cognome_nome;
  TString16					_finesospensione;

	static bool filter_func_sbloccosospesi(const TRelation* rel);

protected:
	virtual bool user_create();
	virtual bool user_destroy();
	virtual bool set_print(int m);
	virtual void set_page(int file, int cnt);
	virtual bool preprocess_page(int file, int counter);
	
public:
  void filtra_sezioni();
  void crea_intestazione();
  void header_sezione(const TString16 codsez, const TString16 codsot);
	TSbloccoSospesi() : _data_sblocco(TODAY), _data_stampa(TODAY), _tiposo(""), _cognome_nome("",25) {}
};

HIDDEN inline TSbloccoSospesi& app() { return (TSbloccoSospesi&) main_app(); }

HIDDEN bool printer_handler(TMask_field& f, KEY k)
{        
  if (k == K_SPACE)
  {                  
    TMask& m = f.mask();
    if (!m.query_mode())
    {
      TString16 config;
      config << "SBLOCCOSO";
  
      TPrinter& p = printer();
      p.set_printtype(normprinter);           // Force configuration update
      p.read_configuration(config);
      if (p.set())
      	f.message_box("Stampante configurata per sblocco sospesi");
    }    
  }
  return TRUE;  
}

void TSbloccoSospesi::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a(LF_SOGGETTI);   
  if (sezini.not_empty())
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

bool TSbloccoSospesi::preprocess_page(int file, int counter)
{ 
	TRectype& recsog = current_cursor()->curr();
	TString80 nome = recsog.get(SOG_COGNOME);
	nome << " ";
	nome << recsog.get(SOG_NOME);
	_cognome_nome = nome;
	if (_definitiva)
	{
		TRectype* key = new TRectype(LF_CONTSAN);
		long codice = recsog.get_long(SOG_CODICE);
		key->put(CON_CODICE, codice);
		int err = _scontrolli->read(key);
		int progcon = _scontrolli->rows()+1;
		TRectype& reccon = _contsan->curr();
		reccon.zero();
		reccon.put(CON_CODICE, codice);
		reccon.put(CON_PROGCON, progcon);
		reccon.put(CON_DATACON, _data_stampa);
		reccon.put(CON_TIPOCON, FINE_SOSPENSIONE);
		_scontrolli->add_row(reccon);
		_scontrolli->rewrite();
		con_reord(recsog, _scontrolli);
		current_cursor()->file().rewrite();
	}	
	// salto pagina se cambio sezione
	const TString16 codsez = recsog.get(SOG_CODSEZ);		
	const TString16 codsot = recsog.get(SOG_CODSOT);		
	if ((_codsez!=codsez)||(_codsot!=codsot))
	{
		if (_codsez != "**")
			printer().formfeed();
		_codsez = codsez;
		_codsot = codsot;
		header_sezione(codsez, codsot);
	}
  return TRUE;
}

void TSbloccoSospesi::set_page(int file, int cnt)
{ 
	set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
	set_row(1,"@9g@S", FLD(LF_SOGGETTI,SOG_CATDON));
	set_row(1,"@12g#a", &_cognome_nome);
	set_row(1,"@38g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
	set_row(1,"@49g@S", FLD(LF_SOGGETTI,SOG_TESSAVIS));
	set_row(1,"@57g@S", FLD(LF_SOGGETTI,SOG_PROS_STATO));
	set_row(1,"@68g@S", FLD(LF_SOGGETTI,SOG_STATO));
	set_row(1,"@80g@ld", FLD(LF_SOGGETTI,SOG_DATA_PROS));
}

bool TSbloccoSospesi::filter_func_sbloccosospesi(const TRelation* rel)
{   
	bool filtrato = FALSE;
	TRectype& recsog = rel->lfile().curr();
	TRectype* key = new TRectype(LF_CONTSAN);
	long codice = recsog.get_long(SOG_CODICE);
	key->put(CON_CODICE, codice);
	int err = app()._scontrolli->read(key);
	if (err == NOERR)
	{   
		TString16 stato = recsog.get(SOG_STATO);
		TDate datastato = recsog.get(SOG_DATASTATO);
		for (int r=1; r<=app()._scontrolli->rows(); r++)
		{                 
			const TRectype& riga = app()._scontrolli->row(r);
			TString16 tipo = riga.get(CON_TIPOCON);
			TDate datatipo = riga.get(CON_DATACON);
			if (tipo == stato && datatipo == datastato)
			{
				TString16 prosstipo = riga.get(CON_PROSSTIPO);
				if (modstato_tcs(prosstipo) == 'F')
				{
					TDate datafine = riga.get(CON_PROSSDATA);
					filtrato = (datafine.ok()) && (datafine <= app()._data_sblocco); 
				}					
			}					
		}
	}	
	return filtrato;
}

void TSbloccoSospesi::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(100);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if ((densot.ok())&& (densot.not_empty()))
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

bool TSbloccoSospesi::set_print(int)
{
  KEY tasto;
  tasto = _msk->run();
	if (tasto == K_ENTER)
	{  	 
		_codsez = "**";
		_codsot = "**";
		_data_sblocco = _msk->get(F_DATA);
		_tiposo = _msk->get(F_TIPO);
		_definitiva = _msk->get_bool(F_DEFINITIVA);
    reset_files(); 
    add_file(LF_SOGGETTI);
		filtra_sezioni();
		// filtro per tipo sospensione                                     
    if (_tiposo.not_empty())
    	current_cursor()->setfilter(format("STATO == \"%s\"",(const char*)_tiposo));
 		else      
      current_cursor()->setfilter("TCS->S6 == \"S\"", TRUE);
		current_cursor()->set_filterfunction(filter_func_sbloccosospesi,TRUE);
		reset_print();		
    crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TSbloccoSospesi::crea_intestazione()
{
	reset_header();

	TString sep(100);
	if (_definitiva)
  	sep = "SBLOCCO DEFINITIVO SOSPESI ";
	else  	
  	sep = "STAMPA PROVVISORIA PER SBLOCCO SOSPESI ";	
  sep << _tiposo;
  sep << " FINO ALLA DATA ";
	TString16 data_stampa = _data_sblocco.string();
	sep << data_stampa;
	sep.center_just();
 	set_header(2, "@0g%s", (const char*) sep);
 	data_stampa = _data_stampa.string();
  set_header(2,"@0g%10s", (const char*) data_stampa);
  sep = "";
  sep << "Pag. @#";
  set_header(2, "@88g%s", (const char*) sep);  
  	
  sep = "";
  sep.fill('-');
  set_header(3, (const char *) sep);
 	set_header(4,"@0gCod.@9gC.@12gCognome e nome@38gNato il@49gTessera@57gStato att.@68gStato prec.@80gScadenza");
 	set_header(5,"@0g------@9g--@12g-------------------------@38g----------@49g-------@57g----------@68g-----------@80g----------");
}

bool TSbloccoSospesi::user_create()
{
	_msk = new TMask("at1300a");
	_rel = new TRelation(LF_SOGGETTI);
	_rel->add(LF_SEZIONI,"CODSEZ==CODSEZ|CODSOT==CODSOT");
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
	//cursore ordinamento per sezione+sottogruppo+cognome+nome  
  _cur = add_cursor(new TCursor(_rel, "", 3));
  _msk->set_handler(F_PRINTER, printer_handler);
	_contsan = new TLocalisamfile(LF_CONTSAN);
	_scontrolli = new TRecord_array(LF_CONTSAN, CON_PROGCON);
	TString16   config;
	config << "SBLOCCOSO";
  printer().read_configuration(config);
	return TRUE;
}	

bool TSbloccoSospesi::user_destroy()	
{
	delete _rel;
	delete _msk;
	delete _contsan;
	delete _scontrolli;
  printer().read_configuration();
	return TRUE;
}

int at1300(int argc, char* argv[])
{
	TSbloccoSospesi a;
	a.run(argc, argv, "Sblocco sospesi");
	return 0;
}