#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <lffiles.h>

#include "soggetti.h"
#include "sezioni.h"
#include <comuni.h>

#include "at4.h"
#include "at4200a.h"

#define ALIAS_LCP	100
#define ALIAS_TCS	200

enum ts { undefined = 0, elenco = 1, etichette = 2 };

// definizione form per etichette
class TUrgenze_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TUrgenze_form(): TForm() {};
  TUrgenze_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TUrgenze_form() {};
};

class TStampaUrgenze : public TPrintapp
{
  static bool filter_func_urgenze(const TRelation* rel);

  TRelation*        _rel;
  TMask*            _msk;
  TUrgenze_form*		_form_eti;
  TRectype*					_sangue;
  TAssoc_array			_categorie;
  long 							_giorni;
  int 							_cur;
  TParagraph_string _cognome_nome, _dencom;
  TDate 						_data_stampa;
  TDate							_dataini, _datafin;
  ts								_tipo_stampa;
  TString16					_codsez, _codsot;
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page (int file, int counter);

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  TMask& app_mask() { return *_msk; }
  TRectype& get_sangue() { return *_sangue; }

  TStampaUrgenze() : _data_stampa(TODAY), _cognome_nome("",25), _dencom("",50) {}
};

HIDDEN inline TStampaUrgenze& app() { return (TStampaUrgenze&) main_app(); }

TCursor* TUrgenze_form::cursor() const { return app().current_cursor(); }

TRelation* TUrgenze_form::relation() const { return cursor()->relation(); }

void TStampaUrgenze::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a(LF_SOGGETTI);   
  if (sezini.not_empty())
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

void TStampaUrgenze::set_page(int file, int cnt)
{
	switch (_tipo_stampa)
	{
		case etichette:
		{
			TPrint_section& corpo = _form_eti->get_body();
			corpo.reset();
			corpo.update();
  		for (int i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			// serve perch alla prossima etichetta rifaccia la setpage
			// altrimenti stampa sempre la stessa etichetta			
			force_setpage(TRUE);
		}
		break;
		case elenco:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
			set_row(1,"@9g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@12g#a", &_cognome_nome);
			set_row(1,"@38g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(1,"@49g@S", FLD(LF_SOGGETTI,SOG_DOM_INDIR));
			set_row(1,"@100g@S", FLD(LF_SOGGETTI,SOG_TELABI));
			set_row(1,"@116g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
			set_row(2,"@49g#a", &_dencom);
			set_row(2,"@116g@S", FLD(LF_SOGGETTI,SOG_TELALT));
			set_row(2,"@38g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
		}
		break;					
	}	
}

bool TStampaUrgenze::filter_func_urgenze(const TRelation* rel)
{   
	bool filtrato = TRUE;
	TLocalisamfile& sog = rel->lfile(); 
	//filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		const TString16 cat = sog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
	}	
	// filtro per	tipizzazione
	if (filtrato)		
	{ 
		TRectype from = app().get_sangue();
  	filtrato = (sog.curr()==from);
	}			
	// filtro su data donazione
	if (filtrato)
	{		
		const long giorni_sez = rel->lfile(LF_SEZIONI).get_long(SEZ_INTMINCONV);
		const long giorni = (giorni_sez > app()._giorni) ? giorni_sez : app()._giorni;
		TDate data(TODAY);
		data = data - (const long) giorni;
		const TDate dataultdon = sog.get(SOG_DATAULTDON);
  	filtrato = (dataultdon <= data);
  }  	
  return filtrato;
}

bool TStampaUrgenze::preprocess_page(int file, int counter)
{
	if (_tipo_stampa == elenco)
	{
		TRectype& recsog = current_cursor()->curr();
		TString80 nome = recsog.get(SOG_COGNOME);
		nome << " ";
		nome << recsog.get(SOG_NOME);
		_cognome_nome = nome;
		TString256 localita = "";
		localita << current_cursor()->curr(-ALIAS_LCP).get("S6");
		if (localita.not_empty() && localita.ok())
		{
			localita << " ";
			localita << current_cursor()->curr(-ALIAS_LCP).get("S0");
			localita << " - ";		
		}
		else
		{
			localita = current_cursor()->curr(LF_COMUNI).get(COM_CAPCOM);
			localita << " ";
		}			
		localita << current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
		localita.trim();
		_dencom = localita;
		// salto pagina se cambio sezione
		const TString16 codsez = recsog.get(SOG_CODSEZ);		
		const TString16 codsot = recsog.get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				printer().formfeed();
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
	}		
  return TRUE;
}

void TStampaUrgenze::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if ((densot.ok())&& (densot.not_empty()))
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

bool TStampaUrgenze::set_print(int)
{
	_tipo_stampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  	_tipo_stampa = elenco;
  	_codsez = "**";
  	_codsot = "**";
  	break;
  case F_ETICHETTE:
  	_tipo_stampa = etichette;	
  	break;
  }	
	if (_tipo_stampa != undefined)
  {
    reset_files(); 
    add_file(LF_SOGGETTI);
		filtra_sezioni();
		// filtro per categorie		
  	const TString16 catpri = _msk->get(F_CAT1);
  	const TString16 catsec = _msk->get(F_CAT2);
  	const TString16 catter = _msk->get(F_CAT3);
  	const TString16 catqua = _msk->get(F_CAT4);
  	const TString16 catqui = _msk->get(F_CAT5);
  	const TString16 catses = _msk->get(F_CAT6);
  	if (catpri.not_empty() && catpri.ok())
  		_categorie.add((const char*) catpri);
  	if (catsec.not_empty() && catsec.ok())
  		_categorie.add((const char*) catsec);
  	if (catter.not_empty() && catter.ok())
  		_categorie.add((const char*) catter);
  	if (catqua.not_empty() && catqua.ok())
  		_categorie.add((const char*) catqua);
  	if (catqui.not_empty() && catqui.ok())
  		_categorie.add((const char*) catqui);
  	if (catses.not_empty() && catses.ok())
  		_categorie.add((const char*) catses);
		// filtro per tipizzazione
  	_sangue->zero();
  	const TString16 gruppoab0 	= _msk->get(F_GRUPPOAB0);
  	const TString16 rhantid 		= _msk->get(F_RHANTID);
  	const TString16 kell 				= _msk->get(F_KELL);
  	const TString16 fenotiporh	=	_msk->get(F_FENOTIPORH);
  	const TString16 du					=	_msk->get(F_DU);
		if ((gruppoab0.ok()) && (gruppoab0.not_empty()))
    	_sangue->put(SOG_GRUPPOAB0, gruppoab0);
		if ((rhantid.ok()) && (rhantid.not_empty()))
    	_sangue->put(SOG_RHANTID, rhantid);
		if ((kell.ok()) && (kell.not_empty()))
    	_sangue->put(SOG_KELL, kell);
		if ((fenotiporh.ok()) && (fenotiporh.not_empty()))
    	_sangue->put(SOG_FENOTIPORH, fenotiporh);
		if ((du.ok()) && (du.not_empty()))
    	_sangue->put(SOG_DU, du);
		_giorni = _msk->get_long(F_GIORNI);
		// filtra solo idonei
    current_cursor()->setfilter("TCS->S6 == \"I\"", TRUE);
    // filtra per categorie, tipizzazione e data donazione
    current_cursor()->set_filterfunction (filter_func_urgenze, TRUE);
		reset_print();
    crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TStampaUrgenze::crea_intestazione()
{
  reset_header();
  if (_tipo_stampa == elenco)
  {
  	TString sep(132);
  	sep = "CONVOCAZIONI URGENTI";
		sep.center_just();
  	set_header(2, "@0g%s", (const char*) sep);
  	TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);
  	sep = "";
  	sep << "Pag. @#";
  	set_header(2, "@120g%s", (const char*) sep);  
  	sep = "";
  	sep.fill('-');
  	set_header(3, (const char *) sep);
  	set_header(4,"@0gCodice@9gC.@12gCognome e nome@38gNato il@49gIndirizzo@100gTelefono abit.@116gTelefono lavoro");
  	set_header(5,"@49gCAP/Localit/Comune/Prov.@116gTelefono altro");
  	set_header(6,"@0g--------@9g--@12g-------------------------@38g----------@49g--------------------------------------------------@100g---------------@116g---------------");
	}  	
}

bool TStampaUrgenze::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCP);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM");
  // per stampare nell'intestazione la denominazione della sezione
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");
	// ordinamento per sezione+sottogruppo+cognome+nome
  _cur = add_cursor(new TCursor(_rel, "", 3));
	_form_eti = new TUrgenze_form("AT_ETSOG");
  _msk = new TMask("at4200a");
  _sangue = new TRectype(LF_SOGGETTI);
  return TRUE;
}

bool TStampaUrgenze::user_destroy()
{
  delete _msk;
  delete _rel;
  delete _form_eti;
  delete _sangue;
  return TRUE;
}

int at4200(int argc, char* argv[])
{
  TStampaUrgenze a;
  a.run(argc, argv, "Urgenze");
  return 0;
}
