#include <printapp.h>
#include <form.h>

#include "mglib.h"
#include "mg3200.h"
#include "mg3frm.h"

class TStampadisp_mask: public TMask
{
  static bool handle_ragg(TMask_field &fld, KEY k); // handler 
  static bool handle_liv(TMask_field &fld, KEY k); // handler 
  static bool handle_subord(TMask_field &fld, KEY k); // handler 
public:
  TStampadisp_mask();
  ~TStampadisp_mask(){};
};
TStampadisp_mask::TStampadisp_mask():
  TMask("mg3200")
{
  TCodgiac_livelli giaclev;
  TCodart_livelli  artlev;
  
  set_handler(F_TOLIVELLOART, handle_liv);
  set_handler(F_TOLIVELLOGIAC, handle_liv);
  set_handler(F_ORDINE, handle_ragg);
  set_handler(F_DETTAGLIODEP, handle_ragg);
  set_handler(F_DETTAGLIOMAG, handle_ragg);
  set_handler(F_ORDINEART,handle_subord);
  if (!artlev.enabled()) {
    TMask_field &f1=field(F_RAGGCODICE);
    f1.reset(); f1.check();f1.hide();
    field(F_FROMLIVELLOART).hide();
    field(F_TOLIVELLOART).hide();
  }
  if (!giaclev.enabled()) {
    TMask_field &f2=field(F_RAGGLIVGIAC);
    f2.reset(); f2.check();f2.hide();
    field(F_FROMLIVELLOGIAC).hide();
    field(F_TOLIVELLOGIAC).hide();
  }
}

bool TStampadisp_mask::handle_subord(TMask_field &fld, KEY k)
{
  if (k == K_SPACE) 
  {
  TOperable_field &fraggcod=(TOperable_field &)fld.mask().field(F_RAGGCODICE);
  TOperable_field &fragggiac=(TOperable_field &)fld.mask().field(F_RAGGLIVGIAC);
  switch (fld.get()[1]) {
    case 'C':
      if (fld.mask().get(F_ORDINE)=="A") 
        fraggcod.enable();
      else {
        if (!fld.mask().get_bool(F_DETTAGLIOMAG) || !fld.mask().get_bool(F_DETTAGLIODEP))
        {
          fraggcod.reset(); fraggcod.disable();
          fragggiac.reset(); fragggiac.disable();
          fraggcod.do_message(0);fragggiac.do_message(0);
        }
      }
    break;
    case 'M':
    case 'D':
      fraggcod.reset(); fraggcod.check();fraggcod.disable();
    break;
    }
  }
  return TRUE;
}

bool TStampadisp_mask::handle_liv(TMask_field &fld, KEY k)
{
  if (k == K_TAB) 
  {
    TOperable_field &ftotmag=(TOperable_field &)fld.mask().field(F_TOTALIMAGAZZINI);
    TOperable_field &fdettmag=(TOperable_field &)fld.mask().field(F_DETTAGLIOMAG);
    if (fld.mask().get(F_ORDINE)=="A") 
    {
      if (fld.mask().get_int(F_TOLIVELLOART) || fld.mask().get_int(F_TOLIVELLOGIAC))
      {
        // raggruppamenti prima del dettaglio magazzini
        ftotmag.reset(); ftotmag.disable();ftotmag.do_message(0);
      } else {
        ftotmag.enable();
      }
    }
  }
  return TRUE;
}

bool TStampadisp_mask::handle_ragg(TMask_field &fld, KEY k)
{
  if (k == K_SPACE) 
  {
    TOperable_field &ftotmag=(TOperable_field &)fld.mask().field(F_TOTALIMAGAZZINI);
    TOperable_field &fraggcod=(TOperable_field &)fld.mask().field(F_RAGGCODICE);
    TOperable_field &fragggiac=(TOperable_field &)fld.mask().field(F_RAGGLIVGIAC);
    if (fld.mask().get(F_ORDINE)=="M")
    {
      ftotmag.enable();
      if (!fld.mask().get_bool(F_DETTAGLIOMAG) || !fld.mask().get_bool(F_DETTAGLIODEP))
      {
        fraggcod.reset(); fraggcod.disable();
        fragggiac.reset(); fragggiac.disable();
        fraggcod.do_message(0);fragggiac.do_message(0);
        return TRUE;
      }
    } 
    fraggcod.enable();
    fragggiac.enable();
  }
  return TRUE;
}


// mg3200       Stampa 
class TStampa_disp : public TPrint_application
{
  TArray * _files;
  TStampadisp_mask * _mask;
  TCursor * _cur;
  TForm_stampemg * _form; // to be moved into TPrint_application

protected:
  virtual bool print_one(int file) 
  {return TRUE;}
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int i=1);
  virtual void set_page(int file,int count)
  { }
  void setprint_permagazzini();
  void setprint_perarticoli();

public:
  TStampa_disp() {}
};


bool TStampa_disp::user_create()
{
  _mask = new TStampadisp_mask;
  _files = new TArray();
  _files->add(new TLocalisamfile(LF_MAG));
  return TRUE;
}

bool TStampa_disp::user_destroy()
{
  delete _mask;
  delete _files;
  return TRUE;
}


void TStampa_disp::setprint_permagazzini()
{
  char subordine=*_mask->get(F_ORDINEART);
  TRectype darec(LF_MAG),arec(LF_MAG);
  TString cfilter,sortexp,joinexp;
  _form = new TForm_stampemg("mg3200b", "");
  _cur = _form->cursor();
  // setta il filtro sul cursore
  if (*_mask->get(F_DAART))
    cfilter << "(CODART>=" <<'"' << _mask->get(F_DAART)<< "\")&&" ;
  if (*_mask->get(F_AART))
    cfilter << "(CODART<=" <<'"' << _mask->get(F_AART)<< "\")&&" ;
  if (cfilter.not_empty())
    cfilter.cut(cfilter.len()-2); 

  // abilita/disabilita le sezioni dei totali
  _form->setdett_permag(_mask->get_int(F_FROMLIVELLOART),
    _mask->get_int(F_TOLIVELLOART),
    _mask->get_int(F_FROMLIVELLOGIAC),
    _mask->get_int(F_TOLIVELLOGIAC),
    TRUE,
    _mask->get_bool(F_TOTALIDEPOSITI),
    _mask->get_bool(F_DETTAGLIODEP));

  darec.put("ANNOES",_mask->get(F_ANNOES));
  darec.put("CODMAG",_mask->get(F_DAMAG));
  arec.put("ANNOES",_mask->get(F_ANNOES));
  arec.put("CODMAG",_mask->get(F_AMAG));
  _cur->setfilter(cfilter);
  _cur->setregion(darec,arec);
}

// ORDINAMENTO 
void TStampa_disp::setprint_perarticoli()
{
  TRectype darec(LF_ANAMAG),arec(LF_ANAMAG);
  TString cfilter,filter,joinexp;
  TString sortexp("ANNOES|CODART|LIVELLO|CODMAG");

  char subordine=*_mask->get(F_ORDINEART);
  _form = new TForm_stampemg("mg3200a", "");
  _cur = _form->cursor();
  if (*_mask->get(F_DAMAG))
    filter << "(CODMAG[1,3]>=" <<'"' << _mask->get(F_DAMAG)<< "\")&&" ;
  if (*_mask->get(F_AMAG))
    filter << "(CODMAG[1,3]<=" <<'"' << _mask->get(F_AMAG)<< "\")&&" ;
  if (filter.not_empty())
    filter.cut(filter.len()-2); 
  switch (subordine)
  {
    case 'C':
      _cur->setkey(1);
      darec.put("CODART",_mask->get(F_DAART));
      arec.put("CODART",_mask->get(F_AART));
      _form->find_field('B',odd_page,"H_CATMER").hide();
      _form->find_field('B',odd_page,"TOT_CATMER").hide();
      _form->setcatmer(FALSE);
    break;
    case 'D':
      _cur->setkey(2);
      darec.put("DESCR",_mask->get(F_DADES));
      arec.put("DESCR",_mask->get(F_ADES));
      _form->find_field('B',odd_page,"H_CATMER").disable();
      _form->find_field('B',odd_page,"TOT_CATMER").hide();
      _form->setcatmer(FALSE);
    break;
    case 'M':
      _cur->setkey(3);
      darec.put("GRMERC",_mask->get(F_DACATMER));
      darec.put("CODART",_mask->get(F_DAART));
      arec.put("GRMERC",_mask->get(F_ACATMER));
      arec.put("CODART",_mask->get(F_AART));
      _form->find_field('B',odd_page,"H_CATMER").show();
      _form->find_field('B',odd_page,"TOT_CATMER").show();
      _form->setcatmer(TRUE);
    break;
  }
  _form->setdett_perart(
    _mask->get_int(F_FROMLIVELLOART),
    _mask->get_int(F_TOLIVELLOART),
    _mask->get_int(F_FROMLIVELLOGIAC),
    _mask->get_int(F_TOLIVELLOGIAC),
    _mask->get_bool(F_TOTALIMAGAZZINI),
    _mask->get_bool(F_DETTAGLIOMAG));


  joinexp << "ANNOES==" << _mask->get(F_ANNOES) << "|CODART==CODART";

  TSortedfile *mag;
  mag= new TSortedfile(LF_MAG,NULL,sortexp,filter,1);
  _cur->relation()->replace(mag,1,joinexp);
  _cur->setfilter(cfilter);
  _cur->setregion(darec,arec);
}


bool TStampa_disp::set_print(int i)
{
  while (TRUE) {
    if (_mask->run() != K_ENTER) 
      return FALSE;

    if (*_mask->get(F_ORDINE)=='A')  
    {
      setprint_perarticoli();
    } else {                      
      setprint_permagazzini();
    }
    _form->print();
    delete _form;
  } // while true  
  return TRUE;
}


int mg3200(int argc, char* argv[])
{
  TStampa_disp a;// derivata da Application e con uso di form
  a.run(argc,argv,"Stampa disponibilit");
  return 0;
}
